from flask import g, render_template, redirect, url_for, request, flash
from sqlalchemy.orm import aliased, joinedload

import mo.db as db
from mo.rights import Right, roles_by_type
from mo.web.jinja import user_url
import mo.users
from mo.web.table import Table, Row, Column
from mo.web import app


@app.route('/org/')
def org_index():
    if 'place' in request.args:
        code = request.args['place']
        place = db.get_place_by_code(code)
        if place is not None:
            return redirect(url_for('org_place', id=place.place_id))
        else:
            flash(f'Místo s kódem {code} neexistuje', 'danger')

    if 'uid' in request.args:
        try:
            uid = int(request.args['uid'])
            user = mo.users.user_by_uid(uid)
            if user is not None:
                return redirect(user_url(user))
            flash(f'Uživatel s ID {uid} neexistuje', 'danger')
        except ValueError:
            flash('ID uživatele musí být číslo', 'danger')

    sess = db.get_session()
    contests_with_role = sess.execute("""
        SELECT contest_id, MIN(role) as role
        FROM contests
            JOIN user_roles ur USING(place_id)
            JOIN rounds r USING(round_id)
            JOIN users USING(user_id)
        WHERE
            user_id = :user_id
            AND r.year = :current_year
            AND (ur.year IS NULL OR ur.year = r.year)
            AND (ur.category IS NULL OR ur.category = r.category)
            AND (ur.seq IS NULL OR ur.seq = r.seq)
        GROUP BY contest_id, r.level, r.category, place_id, r.seq, r.part
        ORDER BY r.level, r.category, place_id, r.seq, r.part
    """, {'user_id': g.user.user_id, 'current_year': mo.current_year})

    contest_ids = []
    roles = []
    for row in contests_with_role:
        contest_ids.append(row['contest_id'])
        roles.append(row['role'])

    contests = (sess.query(db.Contest)
            .select_from(db.Contest)
            .join(db.Round)
            .filter(db.Contest.contest_id.in_(contest_ids))
            .options(joinedload(db.Contest.place))
            .order_by(db.Round.level, db.Round.category, db.Contest.place_id,
                db.Round.seq, db.Round.part)
            .all())

    return render_template('org_index.html', contests=zip(contests, roles), Right=Right,
            role_names=db.role_type_names, gatekeeper=g.gatekeeper)


school_export_columns = (
    Column(key='code', name='kod'),
    Column(key='name', name='nazev'),
    Column(key='town_code', name='kod_obce'),
    Column(key='town', name='obec'),
    Column(key='red_izo', name='red_izo'),
    Column(key='ico', name='ico'),
    Column(key='official_name', name='ofic_nazev'),
    Column(key='address', name='adresa'),
    Column(key='is_zs', name='typ_zs'),
    Column(key='is_ss', name='typ_ss'),
)


@app.route('/org/export/skoly')
def org_export_skoly():
    sess = db.get_session()
    format = request.args.get('format', 'en_csv')

    def gen_rows():
        town = aliased(db.Place)
        for p, s, t in (
                sess.query(db.Place, db.School, town)
                .filter(db.Place.type == db.PlaceType.school)
                .filter(db.Place.place_id == db.School.place_id)
                .filter(db.Place.parent == town.place_id)
                .yield_per(100)):
            yield Row(keys={
                'code': p.get_code(),
                'name': p.name,
                'red_izo': s.red_izo,
                'ico': s.ico,
                'official_name': s.official_name,
                'address': s.address,
                'is_zs': int(s.is_zs),
                'is_ss': int(s.is_ss),
                'town_code': t.get_code(),
                'town': t.name,
            })

    table = Table(school_export_columns, gen_rows(), 'skoly')
    return table.send_as(format, streaming=True)
